From iris.proofmode Require Import tactics.
From iris.bi Require Import fractional.
From iris.algebra Require Import frac.
From lrust.lifetime Require Export lifetime.
From iris.prelude Require Import options.

Class frac_borG Σ := frac_borG_inG : inG Σ fracR.
#[global] Existing Instance frac_borG_inG.

Local Definition frac_bor_inv `{!invGS Σ, !lftG Σ Lat userE, !frac_borG Σ}
                              (φ : Qp → monPred _ _) (V0 : Lat ) γ κ : monPred _ _:=
  (∃ (qfresh : Qp) (qdep qused : option Qp),
      (* We divide the 1 fraction into 3 parts... *)
      ⌜(qfresh ⋅? qdep ⋅? qused = 1)%Qp⌝ ∗
      (* First part: the piece of φ' which is still fresh. *)
      ⎡φ qfresh V0⎤ ∗
      (* Second part: some tokens have been left in deposit while the
         fractured borrow is accessed. *)
      ⎡from_option (λ qdep', qdep'.[κ]) True qdep V0⎤ ∗
      (* Third part: the piece of φ' that is "already used". It can be given
         back when the lifetime ends, but cannot be used when accessing the
         fractured borrow.  *)
      from_option φ True%I qused ∗
      (* We keep a set of "receipts", to prove that we will always have
         enough fractions of the lifetime in deposit. *)
      ⎡own γ (qfresh ⋅? qused)⎤)%I.

Definition frac_bor `{!invGS Σ, !lftG Σ Lat userE, !frac_borG Σ} κ
        (φ : Qp → monPred _ _) : monPred _ _ :=
  (* First, we close up to a bunch of things: iff, lifetime inclusion and view
     inclusion. We also existentially bind γ, the ghost name of receipts. *)
  (∃ γ κ' V0 (φ' : Qp → monPred _ _), κ ⊑ κ' ∗ monPred_in V0 ∗
    ▷ □ (∀ q, φ' q ↔ φ q) ∗
    ▷ □ ⎡(∀ q1 q2, φ' (q1+q2)%Qp ↔ φ' q1 ∗ φ' q2) V0⎤  ∗

    (* Then we use an (internal) atomic persistent borrow. *)
    &in_at{κ'} (frac_bor_inv φ' V0 γ κ'))%I.

Notation "&frac{ κ }" := (frac_bor κ) (format "&frac{ κ }") : bi_scope.

Section frac_bor.
  Context `{!invGS Σ, !lftG Σ Lat userE, !frac_borG Σ}.
  Implicit Types E : coPset.

  Global Instance frac_bor_contractive κ n :
    Proper (pointwise_relation _ (dist_later n) ==> dist n) (frac_bor κ).
  Proof. rewrite /frac_bor /frac_bor_inv. solve_contractive. Qed.
  Global Instance frac_bor_ne κ n :
    Proper (pointwise_relation _ (dist n) ==> dist n) (frac_bor κ).
  Proof. rewrite /frac_bor /frac_bor_inv. solve_proper. Qed.
  Global Instance frac_bor_proper κ :
    Proper (pointwise_relation _ (⊣⊢) ==> (⊣⊢)) (frac_bor κ).
  Proof. rewrite /frac_bor /frac_bor_inv. solve_proper. Qed.

  Lemma frac_bor_iff κ φ φ'  :
    ▷ □ (∀ q, φ q ↔ φ' q) -∗ &frac{κ}φ -∗ &frac{κ}φ'.
  Proof.
    iIntros "#Hφφ' H". iDestruct "H" as (γ κ' V0 φ0) "(? & ? & #Hφ0φ & ?)".
    iExists γ, κ', V0, φ0. iFrame. iNext. iModIntro. iIntros (q).
    iSplit; iIntros "H".
    - iApply "Hφφ'". by iApply "Hφ0φ".
    - iApply "Hφ0φ". by iApply "Hφφ'".
  Qed.

  Global Instance frac_bor_persistent φ κ : Persistent (&frac{κ}φ) := _.

  Lemma bor_fracture' `{LatBottom Lat} φ E κ :
    ↑lftN ⊆ E → ⎡lft_ctx⎤ -∗
      ▷ □ (∀ q1 q2, φ (q1+q2)%Qp ↔ φ q1 ∗ φ q2) -∗ &{κ}(φ 1%Qp) ={E}=∗ &frac{κ}φ.
  Proof.
    iIntros (?) "#LFT #Hfrac Hφ". iMod (own_alloc 1%Qp) as (γ) "Hown". { done. }
    iMod (bor_acc_atomic_strong with "LFT Hφ") as "[H|H]". { done. }
    - iDestruct "H" as (P' κ') "(#Hκκ' & Hφ & Hclose)". iCombine "Hfrac Hφ" as "HH".
      iDestruct (monPred_in_intro with "HH") as (V) "(#HV & #Hfrac' & Hφ)".
      iMod ("Hclose" with "[] [-]") as "Hφ"; last first.
      { iExists γ, κ', V, φ. iFrame "#". iSplitR; [|iSplitR].
        - iIntros "!>!>!>*". by iApply (bi.iff_refl True%I).
        - iModIntro. iNext. iModIntro. iIntros (??).
          rewrite !bi.intuitionistically_elim. iApply ("Hfrac'" $! _ _).
        - by iApply (in_at_bor_share with "[Hφ]"). }
      { iNext. iSplit. { iDestruct "Hφ" as "[$ _]". }
        iExists 1%Qp, None, None; simpl.
        iDestruct "Hφ" as "[_ Hφ]". repeat (iSplit; [done|]). by iFrame. }
      iIntros "!> Hφ H† !>". iNext.
      iDestruct "Hφ" as (q q' q'' Hsum) "(Hφ1 & Htok & Hφ2 & _)".
      destruct q' as [q'|]; simplify_eq/=.
      { iExFalso. rewrite -(embed_pure False). iModIntro ⎡_⎤%I.
        iApply (lft_tok_dead_subj with "Htok [H†]"). iApply "H†". }
      destruct q'' as [q''|]; simplify_eq/=.
      { iDestruct ("Hfrac" with "[Hφ1 Hφ2]") as "H".
        - iSplitL "Hφ1"; [|done]. by iApply monPred_in_elim.
        - by iEval (rewrite -Hsum). }
    by iApply monPred_in_elim.
    - iDestruct (monPred_in_intro with "Hfrac") as (V) "[HV Hfrac']".
      iDestruct "H" as (κ') "(#? & #? & >_)". iExists γ, κ', V, φ. iFrame "#".
      iSplitR; [|iSplitR].
      + iIntros "!> !> !> *". by iApply (bi.iff_refl True%I).
      + iModIntro. iNext. iModIntro. iIntros (??).
        rewrite !bi.intuitionistically_elim. iApply ("Hfrac'" $! _ _).
      + by iApply in_at_bor_fake.
  Qed.
  Lemma bor_fracture `{LatBottom Lat} P φ `{!AsFractional P φ 1} E κ :
    ↑lftN ⊆ E → ⎡lft_ctx⎤ -∗ &{κ}P ={E}=∗ &frac{κ}φ.
  Proof.
    iIntros (?) "#? ?". rewrite [P]as_fractional. iApply bor_fracture'=>//.
    iIntros "!> !> %%".
    assert (Fractional φ). { by apply : as_fractional_fractional. }
    by rewrite fractional -(bi.iff_refl True%I).
  Qed.

  Lemma frac_bor_acc `{LatBottom Lat} φ E q κ :
    ↑lftN ⊆ E →
    ⎡lft_ctx⎤ -∗ &frac{κ}φ -∗ q.[κ] ={E}=∗ ∃ q', ▷ φ q' ∗ (▷ φ q' ={E}=∗ q.[κ]).
  Proof.
    iIntros (?) "#LFT Hfrac Hκ".
    iDestruct "Hfrac" as (γ κ' V0 φ') "#(Hκκ' & HV0 & Hφ' & Hfrac & Hshr)".
    iMod (lft_incl_acc with "Hκκ' Hκ") as (qκ') "[Hκ Hclose]". { done. }
    rewrite -{1}(Qp.div_2 qκ'). rewrite (bi.wand_entails _ _ (lft_tok_split_obj _ _ _)).
    iDestruct "Hκ" as "[Hκ1 Hκ2]".
    iMod (in_at_bor_acc with "LFT Hshr Hκ1") as (Vb) "[H Hclose']"; try done.
    iCombine "H HV0" as "HH".
    iDestruct (monPred_in_intro with "HH") as (V) "(#HV & H & %)".
    iAssert ⎡monPred_in Vb (Vb ⊔ V)⎤%I as "HVVb"; [solve_lat|].
    iDestruct ("H" with "HVVb")
      as (qφ0 qtok qφ1) "(>% & Hφ0 & >Htok' & Hφ1 & >Hown)".
    destruct (Qp.lower_bound (qκ'/2) (qφ0/2)) as (qq & qκ'0 & qφ0' & Hqκ' & Hqφ).
    iExists qq.
    iAssert (▷ φ qq ∗ ⎡▷ φ' (qφ0' + qφ0 / 2) V0⎤)%Qp%I with "[Hφ0]" as "[$ Hqφ]".
    { rewrite -{1}(Qp.div_2 qφ0) {1}Hqφ -assoc_L !embed_later -bi.later_sep. iNext.
      iDestruct ("Hfrac" with "Hφ0") as "[Hφ0 $]". iApply "Hφ'".
      by iApply (monPred_in_elim with "HV0"). }
    iAssert (▷ ⎡own γ qq⎤ ∗
             ▷ ⎡own γ (qφ0' + qφ0 / 2 ⋅? qφ1)%Qp⎤)%I with "[Hown]" as "[>Hown1 Hown2]".
    { rewrite -bi.later_sep -embed_sep -own_op. iNext.
      by rewrite -!frac_op -!cmra_op_opM_assoc_L assoc_L !frac_op -Hqφ Qp.div_2. }
    rewrite monPred_objectively_unfold Hqκ'. iDestruct ("Hκ2" $! V0) as "[Hκ2 Hκ2']".
    iMod ("Hclose'" with "[-Hown1 Hclose Hκ2']") as "Hκ1".
    { iIntros "#HVb". iExists (qφ0' + qφ0 / 2)%Qp, (qtok ⋅ Some qq)%Qp, qφ1.
      iNext. iFrame. iSplit; [|iSplitR "Hφ1"].
      - iPureIntro.
        rewrite (comm_L _ qtok) cmra_opM_opM_assoc_L -assoc_L Some_op_opM /=.
        rewrite -cmra_op_opM_assoc_L frac_op -(comm_L _ qq) assoc_L -Hqφ Qp.div_2.
        by rewrite -cmra_opM_opM_assoc_L.
      - rewrite comm_L Some_op_opM. destruct qtok as [qtok|]; simpl; [|done].
        by iSplitL "Hκ2".
      - destruct qφ1 as [qφ1|]; [|done]. iCombine "HV HVb" as "HH".
        iDestruct (monPred_in_intro with "HH") as (VV) "(HVV & % & %)".
        iApply (monPred_in_elim with "HVV"). iFrame. }
    rewrite -Hqκ'. iClear "Hκκ' HV HVVb". clear dependent qφ0' qtok qφ1 Vb qφ0 V.
    iIntros "!> Hφ". iApply ("Hclose" with "[>-]"). clear κ.
    iMod (in_at_bor_acc with "LFT Hshr Hκ1") as (Vb) "[H Hclose']"; [done|].
    iCombine "H HV0" as "HH".
    iDestruct (monPred_in_intro with "HH") as (V) "(#HV & H & %)".
    iAssert ⎡monPred_in Vb (Vb ⊔ V)⎤%I as "HVVb"; [solve_lat|].
    iDestruct ("H" with "HVVb")
      as (qφ0 qtok qφ1) "(>% & Hφ0 & >Htok' & Hφ1 & >Hown)".
    iEval (rewrite -(Qp.div_2 qκ') {1}Hqκ').
    iClear "Hshr".
    iCombine "Hown1" "Hown" as "Hown". iDestruct (own_valid with "Hown") as %Hval.
    destruct qtok as [qtok|]; simplify_eq/=; last first.
    { destruct (Qp.not_add_le_r qq 1). by rewrite {1}(_ : 1 = qφ0 ⋅? qφ1)%Qp. }
    assert (∃ qm, qtok = qq ⋅? qm)%Qp as [qm ->].
    { assert (qq ≤ qtok)%Qp as [[qm ->]%Qp.lt_sum| ->]%Qp.le_lteq.
      - rewrite (Qp.add_le_mono_r _ _ (qφ0 ⋅? qφ1)). trans 1%Qp; [done|].
        apply reflexive_eq. by rewrite -frac_op -cmra_op_opM_assoc_L comm_L.
      - by exists (Some qm).
      - by exists None. }
    iAssert (⎡from_option (λ qm', qm'.[κ']) True qm V0⎤ ∗ ⎡ monPred_at qq.[κ'] V0 ⎤)%I
      with "[Htok']" as "[Htok1 Htok2]".
    { destruct qm as [qm|]; simplify_eq/=; [|by auto].
      iDestruct "Htok'" as "[$$]". }
    iCombine "Htok2 Hκ2'" as "Htok2"; rewrite -monPred_at_sep -lft_tok_fractional.
    iDestruct (monPred_in_elim with "HV0 Htok2") as "$".
    iApply "Hclose'". iIntros "#HVb". iExists qφ0, qm, (Some (qq ⋅? qφ1)); simpl.
    iNext. iFrame "Hφ0 %". iSplit; [|iSplitL "Htok1"; [done|iSplitR "Hown"]].
    - iPureIntro. by rewrite -cmra_op_opM_assoc_L -(comm_L _ qq)
        -cmra_op_opM_assoc_L (comm_L _ qq) cmra_op_opM_assoc_L.
    - destruct qφ1 as [qφ1|]; simpl; [|by iApply "Hφ'"].
      rewrite !bi.intuitionistically_elim.
      iDestruct (monPred_in_elim with "HV0 Hfrac") as "Hfrac'".
      iApply "Hfrac'". iSplitL "Hφ"; [by iApply "Hφ'"|]. iCombine "HV HVb" as "HH".
      iDestruct (monPred_in_intro with "HH") as (VV) "(HVV & % & %)".
      iApply (monPred_in_elim with "HVV"). iFrame.
    - by rewrite -cmra_op_opM_assoc_L -(comm_L _ qq) cmra_op_opM_assoc_L.
  Qed.

  Lemma frac_bor_shorten φ κ κ' : κ ⊑ κ' -∗ &frac{κ'}φ -∗ &frac{κ}φ.
  Proof.
    iIntros "Hκκ' H". iDestruct "H" as (γ κ0 V0 φ') "[H ?]".
    iExists γ, κ0, V0, φ'. iFrame. by iApply (lft_incl_trans with "Hκκ'").
  Qed.

  Lemma frac_bor_lft_incl `{LatBottom Lat} κ κ' q:
    ⎡lft_ctx⎤ -∗ &frac{κ}(λ q', (q * q').[κ']) -∗ κ ⊑ κ'.
  Proof.
    iIntros "#LFT#Hbor". iApply lft_incl_intro. iModIntro. iIntros (q') "Hκ'".
    iMod (frac_bor_acc with "LFT Hbor Hκ'") as (q'') "[>Htok Hclose]". { done. }
    iExists _. iIntros "{$Htok} !> Hκ'". by iApply "Hclose".
  Qed.
End frac_bor.

Global Typeclasses Opaque frac_bor.
